// AVR HV Programmer Function
// Eleken 2018
// rev_c1815@yahoo.co.jp
// http://eleken.y-lab.org/

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>  /* for sei(), timer0 */
#include <util/delay.h>     /* for _delay_ms() */
#include <avr/eeprom.h>

#include <avr/pgmspace.h>   /* required by usbdrv.h */

#include "writer.h"
#include "sreg.h"
#include "main.h"

static uint8_t HVS_Transfer(uint8_t inst, uint8_t data)
{
	uint8_t i;
	uint8_t ret = 0;
	// first 8 pulses
	for(i=0; i<8; i++){
		SBI(PORTB, PB2);
		_delay_us(1);
		CBI(PORTB, PB2);
		(inst & 0x80)? SBI(PORTB, PB0): CBI(PORTB, PB0);
		(data & 0x80)? SBI(PORTB, PB1): CBI(PORTB, PB1);
		ret = (ret << 1) | ((PIND >> PD5) & 0x01);
		inst = inst << 1;
		data = data << 1;
		_delay_us(1);
	}
	// last 3 pulses
	SBI(PORTB, PB2);
	_delay_us(1);
	CBI(PORTB, PB2);
	
	// clear all pins
	CBI(PORTB, PB1);
	CBI(PORTB, PB0);
	_delay_us(1);
	
	for(i=0; i<2; i++){
		SBI(PORTB, PB2);
		_delay_us(1);
		CBI(PORTB, PB2);
		_delay_us(1);
	}
	
	return ret;
}

static void SetTargetPower(uint8_t mode1, uint8_t mode2)
{
	uint8_t i;
	
	// Power Off
	CBI(PORTD, PD7);
	_delay_ms(1);
	SBI(PORTD, PD6);
	_delay_ms(10);
	
	if(mode1 == HVS_ON){
		CBI(PORTD, PD5);
		SBI(DDRD, PD5); // Output L to SDO pin
		_delay_ms(1);
		
		if(mode2){ // power up mode
			CBI(PORTD, PD6);
			SBI(PORTD, PD7);
		}else{
			CBI(PORTD, PD6);
			_delay_us(50);
			for(i=0; i<6; i++){
				PORTB ^= (1<<PB2);
				_delay_us(1);
			}
			SBI(PORTD, PD7);
			_delay_ms(1);
		}
		
	}else if(mode1 == HVP_ON){
		CBI(PORTD, PD6); // Power ON
		_delay_us(50);
		
		if(mode2 == 0){
			for(i=0; i<6; i++){
				PORTB ^= (1<<PB2); // XTAL1
				_delay_us(1);
			}
		}
		SregSend( 0x20 ); // OE = H
		_delay_us(1);
		SBI(PORTD, PD7); // Vpp
		_delay_us(50);
		SregSend( 0x60 ); // OE,WR = H
	}
	
	_delay_ms(10);
	CBI(DDRD, PD5); // Set SDO to input
}

static void SetXABS(uint8_t mode)
{
	uint8_t temp = SregGetState() & 0xF0;
	SregSend( temp | (mode & 0x0F)); // XA0, XA1, BS1, BS2
}

static void PP_SetByte(uint8_t mode, uint8_t data)
{
	uint8_t temp;
	
	// Set data reg
	PORTC = data >> 2;
	PORTB = (PORTB & 0x0F) | (data << 4);
	
	// Set D[7:0] to output
	DDRC = 0x3F;
	DDRB = (DDRB & 0x0F) | 0x30;
	// Set BS2, BS1, XA1, XA0 pins
	
	SregSend( 0x60 ); // WR, OE = H
	SetXABS(mode);
	
	temp = SregGetState();
	temp |= 0x10; // XT1 = H
	SregSend(temp);
	
	_delay_us(1);

	temp = SregGetState();
	temp = (temp & ~0x18); // XT1 = L, BS2 = L
	SregSend(temp);

	// Set D[7:0] to input
	DDRC = 0;
	DDRB = (DDRB & 0x0F);
	
}

static uint8_t PP_RcvByte(uint8_t mode)
{
	SregSend( 0x40); // WR = H, OE = L
	SetXABS(mode);
	
	_delay_us(10);
	uint8_t ret = (PINC << 2) | ((PINB >> 4) & 0x03);
	
	uint8_t temp = SregGetState();
	temp = (temp | 0x20) & ~0x08; // OE = H, BS2 = L
	SregSend(temp);
	
	return ret;
}
static void PP_StrobeWrite(uint8_t mode, uint8_t delay)
{
	SregSend( 0x20 ); // WR = L, OE = H
	SetXABS(mode);
	_delay_us(10);
	while(delay--){
		_delay_ms(1);
	}
	
	uint8_t temp = SregGetState();
	temp = (temp | 0x40) & ~0x08; // WR = H, BS2 = L
	SregSend(temp);
	
}

// Note: ReadMem() supports reading data from addr 0 sequencially; not supports random address read
// Normally read 40 bytes at once
static uint8_t ReadMem(uint8_t command, uint8_t mode1, uint8_t mode2, uint8_t* buf)
{
	static uint32_t curAddr = 0;
	
	uint8_t readCnt = mode1;
	if(readCnt == 0){
		readCnt = 60;
		curAddr = 0; // reset address
	}
	uint8_t i;
	for(i=0; i<readCnt; i++){
		if(command == RFL_HVS || command == RFL_PP)
		{
			if(curAddr == 0x00) { // Instr. 1
				if(command == RFL_HVS){
					HVS_Transfer(I_LDCMD, C_RD_PRG);
				}else{
					PP_SetByte(XA_1, C_RD_PRG);
				}
			}
			
			if((mode2 == 1) && ((curAddr % (65536UL*2)) == 0)){ // need 24-byte address?
				if(command != RFL_HVS){
					PP_SetByte(BS_2, curAddr >> 17);
				}
			}
			
			if((curAddr % 512) == 0){ // Instr. 2 (send every 512 byte)
				if(command == RFL_HVS){
					HVS_Transfer(I_LDAH, curAddr >> 9);
				}else{
					PP_SetByte(BS_1, curAddr >> 9);
				}
			}
			
			if(curAddr & 1){ // Odd (High Byte)
				if(command == RFL_HVS){
					HVS_Transfer(I_RDLH1, 0x00);
					buf[i] = HVS_Transfer(I_RDLH2, 0x00);
				}else{
					buf[i] = PP_RcvByte(BS_1);
				}
			}else{ // Even (Low Byte)
				if(command == RFL_HVS){
					HVS_Transfer(I_LDAL, curAddr >> 1);
					
					HVS_Transfer(I_RDLL1, 0x00);
					buf[i] = HVS_Transfer(I_RDLL2, 0x00);
				}else{
					PP_SetByte(0, curAddr >> 1);
					buf[i] = PP_RcvByte(0);
				}
			}
		}
		else /*if(command == REE_HVS)*/
		{
			if(curAddr == 0x00){ // Instr. 1
				if(command == REE_HVS){
					HVS_Transfer(I_LDCMD, C_RD_EEP);
				}else{
					PP_SetByte(XA_1, C_RD_EEP);
				}
			}
			if((curAddr % 256) == 0){ // Instr. 2 (send every 256 byte)
				if(command == REE_HVS){
					HVS_Transfer(I_LDAH, curAddr >> 8);
				}else{
					PP_SetByte(BS_1, curAddr >> 8);
				}
			}
			if(command == REE_HVS){
				HVS_Transfer(I_LDAL, curAddr);
				
				HVS_Transfer(I_RDLL1, 0x00);
				buf[i] = HVS_Transfer(I_RDLL2, 0x00);
			}else{
				PP_SetByte(0, curAddr);
				buf[i] = PP_RcvByte(0);
			}
			
		}
		
		curAddr ++;
	}
	
	return 0;
}

// Normally read 32 bytes at once so that blockwise call is guaranteed
static uint8_t WriteMem(uint8_t command, uint8_t mode1, uint8_t mode2, uint32_t addr, const uint8_t* buf)
{
	uint8_t i;
	uint32_t curAddr = addr;
	uint8_t blockSize = mode2; // used only eeprom mode
	
	
	if(command == WFL_HVS || command == WFL_PP)
	{
		if(mode1 & 0x04){ // this is the first block
			if(command == WFL_HVS){
				HVS_Transfer(I_LDCMD, C_WR_PRG);
			}else{
				PP_SetByte(XA_1, C_WR_PRG);
			}
		}
		
		for(i=0; i<16; i++){
			if(command == WFL_HVS){
				HVS_Transfer(I_LDAL, curAddr >> 1);
				HVS_Transfer(I_LDDL, buf[2*i]);
				
				HVS_Transfer(I_LDDH, buf[2*i+1]);
				HVS_Transfer(I_PSTH1, 0);
				HVS_Transfer(I_PSTH2, 0);
			}else{
				PP_SetByte(0, curAddr >> 1);
				PP_SetByte(XA_0, buf[2*i]);
				PP_SetByte(XA_0 | BS_1, buf[2*i+1]);

				uint8_t temp = SregGetState();
				SregSend(temp | 0x80); // PAGEL = H
				_delay_us(1);
				SregSend(temp);
			}
			curAddr += 2;
		}
		if(mode1 & 0x02){ // this is the last block
			if(command == WFL_HVS){
				HVS_Transfer(I_LDAH, addr >> 9);
				HVS_Transfer(I_WRLL1, 0);
				HVS_Transfer(I_WRLL2, 0);
			}else{
				if(mode1 & FL_LARGE) {
					PP_SetByte(BS_2, addr >> 17);
				}
				PP_SetByte(BS_1, addr >> 9);
				PP_StrobeWrite(0, 0);
			}
		}
	}
	else
	{
		if(command == WEE_HVS){
			HVS_Transfer(I_LDCMD, C_WR_EEP);
		}else{
			PP_SetByte(XA_1, C_WR_EEP);
		}
		for(i=0; i<blockSize; i++){
			if(command == WEE_HVS){
				HVS_Transfer(I_LDAL, curAddr);
				HVS_Transfer(I_LDDL, buf[i]);
				HVS_Transfer(I_PSTL1, 0);
				HVS_Transfer(I_PSTL2, 0);
			}else{
				PP_SetByte(0, curAddr);
				PP_SetByte(XA_0, buf[i]);
				
				uint8_t temp = SregGetState();
				SregSend(temp | 0x80); // PAGEL = H
				_delay_us(1);
				SregSend(temp);
			}
			curAddr ++;
		}
		
		if(command == WEE_HVS){
			HVS_Transfer(I_LDAH, curAddr >> 8);
			HVS_Transfer(I_WRLL1, 0);
			HVS_Transfer(I_WRLL2, 0);
		}else{
			PP_SetByte(BS_1, addr >> 8);
			PP_StrobeWrite(0, 0);
		}
	}
	
	curAddr = addr; // store
	return 0;
}

// Pin Map:
// SDO: PD5
// /POW_EN: PD6
// VPP_EN: PD7
// SII: PB0
// SDI: PB1
// SCI: PB2
// D[1:0]: PB[5:4]
// D[7:2]: PC[5:0]

uint8_t WriterFunc(const uint8_t* rcvBuf, uint8_t* sendBuf)
{
	uint8_t msgID = rcvBuf[3];
	sendBuf[0] = (PIND& (1<<PD5))? 0x00: 0x80; // ready/busy: 0x80 when PD5 = L; 
	sendBuf[1] = 0; // rcv_data
	sendBuf[2] = 0; // hvs_data
	sendBuf[3] = msgID;
	
	switch(rcvBuf[0]){
		/*
		// unused...
	case LPT_DAT:
		break;
	case LPT_STA:
		break;
	case LPT_CTL:
		break;
	case USB_DAT:
		break;*/
		
	case HVS_DAT:
		sendBuf[2] = HVS_Transfer(rcvBuf[1], rcvBuf[2]);
		break;
	case USB_POW:
		SetTargetPower(rcvBuf[1], rcvBuf[2]);
		break;
		// Only used for parallel programming...
	case SET_BYTE:
		PP_SetByte(rcvBuf[1], rcvBuf[2]);
		break;
	case RCV_BYTE:
		sendBuf[1] = PP_RcvByte(rcvBuf[1]);
		break;
	case STB_PAGEL:
		{
			uint8_t temp = SregGetState();
			SregSend(temp | 0x80); // PAGEL = H
			_delay_us(1);
			SregSend(temp);
		}
		break;
	case STB_WR:
		PP_StrobeWrite(rcvBuf[1], rcvBuf[2]);
		break;
		
		// Read Flash
	case RFL_PP:
	case REE_PP:
	case RFL_HVS:
	case REE_HVS:
		ReadMem(rcvBuf[0], rcvBuf[1], rcvBuf[2], &sendBuf[4]);
		break;

		// Erase and Write Flash
	case WFL_PP:
	case WEE_PP:
	case WFL_HVS:
	case WEE_HVS:
		{
			uint32_t addr = (
				(((uint32_t)rcvBuf[4]) << 24) | 
				(((uint32_t)rcvBuf[5]) << 16) | 
				(((uint32_t)rcvBuf[6]) << 8) | 
				(((uint32_t)rcvBuf[7])) );
			WriteMem(rcvBuf[0], rcvBuf[1], rcvBuf[2], addr, &rcvBuf[8]);
		}
		break;
	default:
		break;
	}
	
	return 0;
}
